#include <QThread>
#include <QTimer>
#include <QSettings>
#include <QTranslator>
#include <QFileDialog>
#include <QDesktopServices>
#include <QSpinBox>
#include <QMenu>
#include <QPainter>

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "mousewheelguard.h"
#include "widgets/tabs/axesconfig.h"
#include "themeswitcher.h"
#include "advancedsettings.h"

#include "common_types.h"
#include "global.h"
#include "deviceconfig.h"
#include "version.h"

#include <QDebug>

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
    , ui(new Ui::MainWindow)
{
    qDebug()<<"main + member initialize time ="<< gEnv.pApp_start_time->elapsed() << "ms";
    QElapsedTimer timer;
    timer.start();

    ui->setupUi(this);

    QMainWindow::setWindowIcon(QIcon(":/Images/icon-32.png"));

    // firmware version
    setWindowTitle(tr("EvilRacing Configurator") + " v" + APP_VERSION);

    // load application config
    loadAppConfig();

    m_thread = new QThread;
    m_hidDeviceWorker = new HidDevice();
    m_threadGetSendConfig = new QThread;

    m_flasherConnected = false;
    m_oldFirmware = false;

    qDebug()<<"before add widgets ="<< timer.restart() << "ms"<<sizeof(gEnv.pDeviceConfig->config.shift_config);
                                            //////////////// ADD WIDGETS ////////////////
    // add axes widget
    m_axesConfig = new AxesConfig(this);
    ui->layoutV_tabAxesConfig->addWidget(m_axesConfig);
    qDebug()<<"axes config load time ="<< timer.restart() << "ms";

                                            //////////////// SIGNASL-SLOTS ////////////////loadDeviceConfigFromFile

    // get/send config
    connect(this, &MainWindow::getConfigDone, this, &MainWindow::configReceived);
    connect(this, &MainWindow::sendConfigDone, this, &MainWindow::configSent);

    // set selected hid device
    connect(ui->comboBox_HidDeviceList, SIGNAL(currentIndexChanged(int)),
                this, SLOT(hidDeviceListChanged(int)));

    // HID worker
    m_hidDeviceWorker->moveToThread(m_thread);
    connect(m_thread, &QThread::started, m_hidDeviceWorker, &HidDevice::processData);

    connect(m_hidDeviceWorker, &HidDevice::paramsPacketReceived, this, &MainWindow::getParamsPacket);
    connect(m_hidDeviceWorker, &HidDevice::deviceConnected, this, &MainWindow::showConnectDeviceInfo);
    connect(m_hidDeviceWorker, &HidDevice::deviceDisconnected, this, &MainWindow::hideConnectDeviceInfo);
    connect(m_hidDeviceWorker, &HidDevice::flasherConnected, this, &MainWindow::flasherConnected);
    connect(m_hidDeviceWorker, &HidDevice::flasherFound, this, [&](bool found){m_flasherConnected = found;});
    connect(m_hidDeviceWorker, &HidDevice::hidDeviceList, this, &MainWindow::hidDeviceList);


    // read config from device
    connect(m_hidDeviceWorker, &HidDevice::configReceived, this, &MainWindow::configReceived);
    // write config to device
    connect(m_hidDeviceWorker, &HidDevice::configSent, this, &MainWindow::configSent);


    // load config //////////////////////////////////////////////////////// add load config when connecting
    loadDefaultConfig();

    m_thread->start();

    qDebug()<<"after widgets load time ="<< timer.elapsed() << "ms";
    qDebug()<<"without style startup time ="<< gEnv.pApp_start_time->elapsed() << "ms";
}

MainWindow::~MainWindow()
{
    saveAppConfig();
    m_hidDeviceWorker->setIsFinish(true);
    m_hidDeviceWorker->deleteLater();
    m_thread->quit();
    m_thread->deleteLater();
    m_thread->wait();
    m_threadGetSendConfig->quit();
    m_threadGetSendConfig->deleteLater();
    m_threadGetSendConfig->wait();
    delete ui;
}




                                                        ///////////////////// device reports /////////////////////
// device connected
void MainWindow::showConnectDeviceInfo()
{
    m_deviceChanged = true;
    if (ui->comboBox_HidDeviceList->count() > 1) {      // junk detail: void MainWindow::getParamsPacket()
        ui->label_DeviceStatus->setText(QString::number(ui->comboBox_HidDeviceList->count()) + " " + tr("devices connected"));
    } else {
        ui->label_DeviceStatus->setText(tr("Сonnected"));
    }
    if (ui->comboBox_HidDeviceList->itemData(ui->comboBox_HidDeviceList->currentIndex()).toInt() != 1) {
//        ui->pushButton_WriteConfig->setEnabled(true);
//        ui->pushButton_ReadConfig->setEnabled(true);
//        m_oldFirmware = false;
//        ui->label_DeviceStatus->setStyleSheet("color: white; background-color: rgb(0, 128, 0);");
    } else {
        // for old firmware
        ui->pushButton_WriteConfig->setEnabled(false);
        ui->pushButton_ReadConfig->setEnabled(false);
        m_oldFirmware = true;
        ui->label_DeviceStatus->setText(tr("Please update your firmware"));
        ui->label_DeviceStatus->setStyleSheet("color: white; background-color: rgb(168, 168, 0);");
        if (m_flasher) {
            m_flasher->deviceConnected(true);
        }
    }
}

// device disconnected
void MainWindow::hideConnectDeviceInfo()
{
    ui->label_DeviceStatus->setText(tr("Disconnected"));
    ui->label_DeviceStatus->setStyleSheet("color: white; background-color: rgb(160, 0, 0);");
    ui->pushButton_ReadConfig->setEnabled(false);
    ui->pushButton_WriteConfig->setEnabled(false);

    if (m_flasher) {
        m_flasher->deviceConnected(false);
    }
}

// flasher connected
void MainWindow::flasherConnected()
{
    ui->label_DeviceStatus->setText(tr("Connected"));
    ui->label_DeviceStatus->setStyleSheet("color: white; background-color: rgb(0, 128, 0);");
    ui->pushButton_ReadConfig->setEnabled(false);
    ui->pushButton_WriteConfig->setEnabled(false);

    if (m_flasher) {
        m_flasher->deviceConnected(false);
    }
}

// add/delete hid devices to/from combobox
void MainWindow::hidDeviceList(const QList<QPair<bool, QString>> &deviceNames)
{
    if (deviceNames.size() == 0) {
        ui->comboBox_HidDeviceList->clear();
        return;
    } else {
        ui->comboBox_HidDeviceList->clear();
        for (int i = 0; i < deviceNames.size(); ++i) {
            if (deviceNames[i].first) {
                // for old firmware
                ui->comboBox_HidDeviceList->addItem(deviceNames[i].second, 1);
            } else {
                ui->comboBox_HidDeviceList->addItem(deviceNames[i].second);
            }
        }
    }
}

// received device report
void MainWindow::getParamsPacket(uint8_t *buffer)
{
    ReportConverter::paramReport(buffer);

    static QElapsedTimer badTimer;      // todo refactoring

    if (badTimer.isValid() == false) {
        badTimer.start();
    }
    else if (badTimer.elapsed() > 150)
    {
        if (m_deviceChanged) {
            if (gEnv.pDeviceConfig->paramsReport.firmware_version == FIRMWARE_VERSION) {
                if (ui->comboBox_HidDeviceList->count() > 1) {
                    ui->label_DeviceStatus->setText(QString::number(ui->comboBox_HidDeviceList->count()) + " " + tr("devices connected"));
                } else {
                    ui->label_DeviceStatus->setText(tr("Сonnected"));
                }
                ui->pushButton_WriteConfig->setEnabled(true);
                ui->pushButton_ReadConfig->setEnabled(true);
                m_oldFirmware = false;
                m_deviceChanged = false;
                ui->label_DeviceStatus->setStyleSheet("color: white; background-color: rgb(0, 128, 0);");
            } else {
                ui->pushButton_WriteConfig->setEnabled(false);
                ui->pushButton_ReadConfig->setEnabled(false);
                m_oldFirmware = true;
                ui->label_DeviceStatus->setText(tr("Please update your firmware"));
                ui->label_DeviceStatus->setStyleSheet("color: white; background-color: rgb(168, 168, 0);");
                if (m_flasher) {
                    m_flasher->deviceConnected(true);
                }
            }
        }
        badTimer.restart();
    }

    static QElapsedTimer timer;
    static bool change = false;

    m_axesConfig->updateButtonsState();

    if (!change)
    {
        timer.start();
        change = true;
    }
    else if (timer.elapsed() > 17)    // обновление раз в 17мс(~60fps), мб сделать дефайн в герцах
    {
        m_axesConfig->axesValueChanged();
        change = false;
    }
}

// Flasher controller
void MainWindow::deviceFlasherController(bool isStartFlash)
{        // херня? mb QtConcurrent::run()
    // так оставить или как read/write а bool через сигнал?
    QEventLoop loop; // static?
    QObject context;
    context.moveToThread(m_threadGetSendConfig);
    connect(m_threadGetSendConfig, &QThread::started, &context, [&]() {
        qDebug()<<"Start flasher controller";
        if (isStartFlash == true){
            qDebug()<<"Start flash";
            m_hidDeviceWorker->flashFirmware(m_flasher->fileArray());       // NOT SAFETY
        } else {
            qDebug()<<"Enter to flash mode";
            m_hidDeviceWorker->enterToFlashMode();
        }
        qDebug()<<"Flasher controller finished";
        loop.quit();
    });
    m_threadGetSendConfig->start();
    loop.exec();
    m_threadGetSendConfig->quit();
    m_threadGetSendConfig->wait();
}


                                            /////////////////////    APP SETTINGS   /////////////////////

// slot interface style changed
void MainWindow::interfaceStyleChanged(bool isDark)
{
    if (isDark) ThemeSwitcher::setDark();
    else ThemeSwitcher::setWhite();
    updateColor();
}

//! QPixmap gray-scale image (an alpha map) to colored QIcon
QIcon MainWindow::pixmapToIcon(QPixmap pixmap, const QColor &color)
{
    // initialize painter to draw on a pixmap and set composition mode
    QPainter painter(&pixmap);
    painter.setCompositionMode(QPainter::CompositionMode_SourceIn);
    // set color
    painter.setBrush(color);
    painter.setPen(color);
    // paint rect
    painter.drawRect(pixmap.rect());
    // here is our new colored icon
    return QIcon(pixmap);
}

void MainWindow::updateColor()
{
    QColor col = QApplication::palette().color(QPalette::Text);
    ui->toolButton_AdvSet->setIcon(pixmapToIcon(QPixmap(":/Images/advSet.png"), col));
}


// load app config
void MainWindow::loadAppConfig()
{
    QSettings *appS = gEnv.pAppSettings;
    qDebug()<<"Loading application config";

    // load window settings
    appS->beginGroup("WindowSettings");
    this->restoreGeometry(appS->value("Geometry").toByteArray());

    m_themeIsDark = appS->value("DarkTheme", false).toBool();
    interfaceStyleChanged(m_themeIsDark);
    appS->endGroup();

    qDebug()<<"Finished loading application config";
}

// save app config
void MainWindow::saveAppConfig()
{
    QSettings *appS = gEnv.pAppSettings;
    qDebug()<<"Saving application config";
    // save window settings
    appS->beginGroup("WindowSettings");
    appS->setValue("Geometry",   this->saveGeometry());
    appS->setValue("DarkTheme", m_themeIsDark);
    appS->endGroup();
    qDebug()<<"done";
}


                                                            ////////////////// buttons //////////////////
// comboBox selected device
void MainWindow::hidDeviceListChanged(int index)
{
    m_hidDeviceWorker->setSelectedDevice(index);
}
// read config from device
void MainWindow::on_pushButton_ReadConfig_clicked()
{
    qDebug()<<"Read config started";
    ui->pushButton_ReadConfig->setEnabled(false);
    ui->pushButton_WriteConfig->setEnabled(false);

    m_hidDeviceWorker->getConfigFromDevice();
}
// write config to device
void MainWindow::on_pushButton_WriteConfig_clicked()
{
    qDebug()<<"Write config started";
    ui->pushButton_WriteConfig->setEnabled(false);  // не успевает блокироваться? таймер
    ui->pushButton_ReadConfig->setEnabled(false);

//    if (m_trickForRgCompatib == false) {
//        m_trickForRgCompatib = true;
//        on_pushButton_ReadConfig_clicked();
//    } else {
        writeToConfig();
        m_hidDeviceWorker->sendConfigToDevice();
//    }
}
// load from file
void MainWindow::on_pushButton_LoadFromFile_clicked()
{
    qDebug()<<"Load from file started";
    QString fileName = QFileDialog::getOpenFileName(this,
                                                    tr("Open Config"), QDir::currentPath() +'/'+ "configs" + "/",
                                                    tr("Config Files (*.ercfg)"));

    QSettings device_settings( fileName, QSettings::IniFormat );
    loadDeviceConfigFromFile(&device_settings);
    qDebug()<<"done";
}
// save to file
void MainWindow::on_pushButton_SaveToFile_clicked()
{
    qDebug()<<"Save to file started";
    QString tmpStr("");
    if (tmpStr == "") {
        tmpStr = gEnv.pDeviceConfig->config.device_name;
    }

    QFileInfo file(QFileDialog::getSaveFileName(this, tr("Save Config"),
                                                QDir::currentPath() +'/'+ "configs" + "/" + tmpStr,
                                                tr("Config Files (*.ercfg)")));
    QSettings device_settings(file.absoluteFilePath(), QSettings::IniFormat);
    saveDeviceConfigToFile(&device_settings);

    qDebug()<<"done";
}
// load default config file
void MainWindow::on_pushButton_LoadDefaultConfig_clicked()
{
    qDebug()<<"Load default config started";
    gEnv.pDeviceConfig->resetConfig();
    readFromConfig();
    qDebug()<<"Load default config finished";
}

// advanced settings
void MainWindow::on_toolButton_AdvSet_clicked()
{
    AdvancedSettings advSettings(this);
    connect(&advSettings, &AdvancedSettings::themeChanged, this, [&](bool check) {
        m_themeIsDark = check;
        interfaceStyleChanged(m_themeIsDark);
    });

    m_flasher = advSettings.flasher();
    if (m_flasherConnected) {
        advSettings.flasherFound(m_flasherConnected);
    } else if (m_oldFirmware) {
        advSettings.unlockEnterFlasherMode(true);

    }
    // device connected
    connect(m_hidDeviceWorker, &HidDevice::deviceConnected, &advSettings, [&](){advSettings.setBlockUi(false);});
    // enter flash mode clicked
    connect(m_flasher, &Flasher::flashModeClicked, this, &MainWindow::deviceFlasherController);
    // flasher found
    connect(m_hidDeviceWorker, &HidDevice::flasherFound, m_flasher, &Flasher::flasherFound);
    // start flash
    connect(m_flasher, &Flasher::startFlash, this, &MainWindow::deviceFlasherController);
    // flash status
    connect(m_hidDeviceWorker, &HidDevice::flashStatus, m_flasher, &Flasher::flashStatus);

    if (advSettings.exec() == QDialog::Accepted) {

    }

    m_flasher = nullptr;
    saveAppConfig();
}
